/* ***************************************
   pocketmark.c - (C) Fernando Carello 1999-2000 (fcarello@tiscalinet.it)

   A benchmark for the Palm Pilot organizer

   As for now, tests are performed to measure:

   - the CPU speed, by executing some integer and floating point operations 
      (like finding prime numbers)
   - the memory speed, by moving around some data blocks
   - the OS speed, by filling and sorting records in a database
   - the graphics speed, by performing a rectangle copy test

Bugs:
=====
   -  Using gccfloats.c, a fatal runtime error sometimes occurs 
      ("trace" invalid instruction): reversed to use plain PalmOS calls;
      the Float test is now slower by a 40%, and the code is bulkier by a 100% :-(
      

Changelog:
==========
   0.0 (11 nov. 1999) - Start of development
   0.1 (15 dec. 1999) - First working version; tested on PalmOS 2 Debug, PalmOS 3 Debug,
                        PalmOS 3.1 Debug, PalmOS 2.0.5, PalmOS 3.3. 
                        On PalmIIIx / OS 3.3 , indexes are all "100"
   0.1.1 (3 jan. 2000)- Typical results table added to Help menu; documentation update
   0.2 (2 feb. 2000)  - Forms reorganized (now uses 3 forms)
                        DBMark added 
                        GraphMark added
                        Save to Memopad added
                        Indexes tuned
                        Results table updated
                        Help strings updated
   0.2.1 (24 may 2000)- Fixed "Read from Low Memory" bug (tip was disabling "Dynamic Size"
                        from results fields)
                        First release under the "full" GPL

Todo:
=====
   0.3                 -  adding of CPU- and OS-dependent greyscale video benchmark
                                 do not require MathLib nor OS 2+, simply skip according benchmarks
                                 source cleanup for distribution
                                 
   0.4                 -  add system info collecting
                                 

     
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    

**************************************** */

#include <Pilot.h>
#define  DONT_USE_FP_TRAPS 1
#include <System/NewFloatMgr.h>
#include <Window.h>
#include <Rect.h>
#include <MathLib.h>
#include <callback.h>
#include "pocketmark.h"

/**********
* Globals *
**********/
UInt        IntMarks, MemMarks, FloatMarks, DBMarks, 
            GraphMarks; 
char        IntMarksString[10], MemMarksString[10], FloatMarksString[10], 
            PalmMarksString[10], DBMarksString[10], GraphMarksString[10];
            // will store test results across the forms
            
DmOpenRef   testDB;
LocalID     testDBID;
ULong       DBType = 'TEST';
ULong       Creator = 'MarK'; // ... I hate #defines !
typedef struct
  {
    char  text[15];
    Int   number;
  } TestRecordType;
/**************
* End globals *
**************/

void *GetObjectPtr (Word objectID)
{
  FormPtr pFrm;
 
  pFrm = FrmGetActiveForm();	// Gets pointer to the current form
  return(FrmGetObjectPtr(pFrm, FrmGetObjectIndex(pFrm, objectID)));
}

void EraseField (Word FieldID)
{
   FormPtr	Frm;
   FieldPtr	CurrentField;

   Frm = FrmGetActiveForm();
   CurrentField = (FieldPtr) GetObjectPtr(FieldID);
   FldFreeMemory (CurrentField);
   FldEraseField (CurrentField);
   FldDrawField (CurrentField);
   FrmSetFocus (Frm,FrmGetObjectIndex(Frm, FieldID));
}


/* *********************
** StartApplication   **
***********************/

static Boolean StartApplication (void)
{
  Err	      Error;
  DWord	   OSVersion;  
  
  Error = FtrGet ('psys', 1, &OSVersion);
  ErrFatalDisplayIf((OSVersion < 0x02003000), "PalmOS 1.x found: this program requires PalmOS 2.0 or later"); 
  Error = SysLibFind(MathLibName, &MathLibRef);
  if (Error)
    Error = SysLibLoad(LibType, MathLibCreator, &MathLibRef);
  ErrFatalDisplayIf(Error, "MathLib Not Found! PocketMark requires PalmOS >= 2.0 and MathLib.prc"); 
  Error = MathLibOpen(MathLibRef, MathLibVersion);
  ErrFatalDisplayIf(Error, "Error opening MathLib.prc, or bad MathLib version! PocketMark requires MathLib"); 

  testDBID = DmFindDatabase (0,"pmark_db");
  if (testDBID)   // A previous launch failed to clear the test database
  {
    Error = DmDeleteDatabase (0, testDBID);    
    ErrFatalDisplayIf (Error, "Can't delete old test DB !");
  }
  Error = DmCreateDatabase (0,"pmark_db",Creator,DBType,false);
  ErrFatalDisplayIf (Error, "Can't create test DB !");
  testDBID = DmFindDatabase (0,"pmark_db");
  testDB = DmOpenDatabase (0,testDBID,dmModeReadWrite);
  ErrFatalDisplayIf ((!testDB), "Can't open test DB !"); // it was impossible to open the DB we just created

  FrmGotoForm(MainForm);         
  return(false);
}

/*********************
** StopApplication  **
*********************/

static void StopApplication (void)
{
  UInt 	   UseCount;
  Err 	   Error;

  Error = MathLibClose(MathLibRef, &UseCount);
  ErrFatalDisplayIf(Error, "Can't close MathLib!");
  if (UseCount == 0)
    SysLibRemove(MathLibRef); 
  Error = DmCloseDatabase (testDB);
  ErrFatalDisplayIf (Error, "Fatal Error: can't close test DB !");
  Error = DmDeleteDatabase (0,testDBID);
  ErrFatalDisplayIf (Error, "Fatal Error: can't delete test DB !");
  FrmCloseAllForms();
}

/************
** IntMark **
************/

UInt IntMark() // Something like a raw algorithm to find perfect numbers: only integer ops are performed
{
  ULong  Number, MaxDivider, Divider, Sum;
  ULong  StartTicks, StopTicks, Duration[3];
  UInt   NPerfects=0, Result, i;
  double DblDuration, DblResult;

  for (i=0; i<=2; i++)  // Better to perform 3 times the test, to have a more stable result
  {  
    StartTicks = TimGetTicks();
    for (Number=3; Number <= 200; Number++)
    {
      MaxDivider = (Number/2);		
      Sum = 1;
      for (Divider=2; Divider <= MaxDivider; Divider++)
        if  ((Number % Divider)==0) Sum += Divider;
      if (Sum==Number) NPerfects++; // Number is perfect
    }
    StopTicks = TimGetTicks();
    Duration[i] = StopTicks - StartTicks;
  }
  DblDuration = (Duration[0] + Duration[1] + Duration[2]) / 3.0;
  MathLibRInt (MathLibRef, (8000.0/DblDuration), &DblResult);
  Result = (UInt) DblResult;

  return (Result);
}

/************
** MemMark **
************/

UInt MemMark() // A memory block mover
{
  ULong     StartTicks, StopTicks;
  UInt      i, j, Result, Duration[3];
  ULongPtr  Dest; 
  double    DblDuration, DblResult;
  
  Dest   = (ULongPtr) MemPtrNew (16000);
  for (i=0; i<=2; i++)
  {
    StartTicks = TimGetTicks();
    for (j=0; j<= 100; j++) MemSet (Dest, 16000, 1);
    StopTicks = TimGetTicks();
    Duration[i] = StopTicks - StartTicks;
  }
  MemPtrFree (Dest);
  DblDuration = (Duration[0] + Duration[1] + Duration[2]) / 3.0;
  MathLibRInt (MathLibRef, (8876.1/DblDuration), &DblResult);
  Result = (UInt) DblResult;

  return(Result);
}

/**************
** FloatMark **
**************/

UInt FloatMark()  // Not very meaningful: just a reasonable bunch of float operations (using doubles)
{
  double    Base=1.0, k1=0.01, k2=-0.011, Temp1=1.02, Temp2=1.0, DblDuration, DblResult;
  ULong     StartTicks, StopTicks;
  UInt      i, j, k, Result, Duration[3];

  CALLBACK_PROLOGUE
  for (i=0; i<=2; i++)
  {
    StartTicks = TimGetTicks();
    for (j=0; j<=4; j++)
    {
      for (k=0; k<=70; k++)
//        Base = __adddf3 (Base,(__muldf3(Temp2,__divdf3(__adddf3(k1,__divdf3(k,100000.0)),__subdf3(k2,__divdf3(k,100003.0))))));
        Base += (Temp2*(k1+k/100000.0)/(k2-k/100003.0));
      MathLibCos (MathLibRef, Base, &Temp1);
      MathLibFAbs (MathLibRef, Temp1, &Temp2);
      Base = Temp2/Temp1;
      MathLibRInt (MathLibRef, Base, &Base);
    }
    StopTicks = TimGetTicks();
    Duration[i] = StopTicks - StartTicks;
  }
  DblDuration = (Duration[0] + Duration[1] + Duration[2]) / 3.0; 
  MathLibRInt (MathLibRef, 8530.0 / DblDuration, &DblResult);
  Result = (UInt) DblResult;  

  CALLBACK_EPILOGUE
  return (Result);
}

/*********
* DBMark *
*********/

/* Let's define the brain-damaged CompareFunction for DmQuickSort... */
static Int CompareFunction 
 (TestRecordType* rec1, TestRecordType* rec2, Int i, SortRecordInfoPtr s1, SortRecordInfoPtr s2, VoidHand appInfoH)  
{
 return (rec1->number - rec2->number);
}

UInt DBMark()  // We try to get an index of both the effectiveness of PalmOS in managing databases, and the memory access speed
{              // First we write some records, then we order them using QuickSort
               // Unfortunately, the performance figure depends on how much the memory is fragmented
               // at the time we perform the test ... need some rework
  ULong           StartTicks, StopTicks, tempDuration;
  UInt            i, Result;
  double          DblDuration, DblResult;
  char            TestText[15];
  TestRecordType  TestRecord;
  TestRecordType  *TestRecordPtr;
  VoidHand        TestRecordHandle;
  UInt            InsertionIndex = 0;  // We add new records on top
  Err             Error = 0;

  StartTicks = TimGetTicks();
  for (i=0; i<=199; i++)
  {
    TestRecord.number = 200-i;
    StrIToA(TestText, i);
    StrCopy (TestRecord.text,TestText);
    TestRecordHandle = DmNewRecord (testDB, &InsertionIndex, sizeof(TestRecordType));
    TestRecordPtr    = MemHandleLock (TestRecordHandle);
    Error = DmWrite (TestRecordPtr, 0, &TestRecord, sizeof(TestRecord));
    MemHandleUnlock (TestRecordHandle);
    DmReleaseRecord (testDB, InsertionIndex, true);
  }
  Error = DmQuickSort(testDB,(DmComparF*)CompareFunction,0);
  StopTicks = TimGetTicks();
  tempDuration = StopTicks - StartTicks;  
  DblDuration = tempDuration / 1.0;
  MathLibRInt (MathLibRef, (28400.0/DblDuration), &DblResult);
  Result = (UInt) DblResult;

  return(Result);
}

/************
* GraphMark *
************/

UInt GraphMark()  // A simple graph block copying test
{
  ULong           StartTicks, StopTicks, tempDuration;
  UInt            x, y;
  double          DblDuration, DblResult;
  WinHandle       DrawWinHandle;  
  RectangleType   DrawWinBounds, ClipBounds;
  Word            WinError;
  
  StartTicks = TimGetTicks();
  RctSetRectangle (&DrawWinBounds, 0, 15, 159, 145); // This is drawing frame  
  RctSetRectangle (&ClipBounds,0,15,32,32); // This is the clipping rectangle (block)
  DrawWinHandle = WinCreateWindow (&DrawWinBounds, 0, false, true, &WinError);
  ErrFatalDisplayIf (WinError, "Can't create window !"); 
  WinEnableWindow (DrawWinHandle);
  WinInitializeWindow (DrawWinHandle);
  WinSetDrawWindow (DrawWinHandle);
  WinEraseWindow();
  WinDrawRectangle (&ClipBounds,0);
  for (x=0; x<=80; x++)
    for (y=0; y<=80; y++)
      WinCopyRectangle (DrawWinHandle, DrawWinHandle, &ClipBounds, x, y, scrCopy);
  WinDeleteWindow (DrawWinHandle,true);  
  StopTicks = TimGetTicks();
  tempDuration = StopTicks - StartTicks;
  DblDuration = tempDuration / 1.0;
  MathLibRInt (MathLibRef, (130900.0/DblDuration), &DblResult);
  return ((UInt) DblResult);
}


/************************
** MainFormHandleEvent **
************************/

static Boolean MainFormHandleEvent (EventPtr event)
{
  FormPtr	frm;
  FieldPtr	ActiveField;
  char      WaitmessageString[100];
  Boolean	handled=false;

  CALLBACK_PROLOGUE  
  switch (event->eType)
  {
    case frmOpenEvent:	   // the form was just opened: draw it
      frm = FrmGetActiveForm();
      FrmDrawForm (frm);
      handled = true;
    break;

    case ctlSelectEvent:	// a button was pressed
      switch (event->data.ctlEnter.controlID)
      {
        case Main_GoButton:
            
          CtlSetEnabled ((ControlPtr)GetObjectPtr (Main_GoButton),false);
          ActiveField = (FieldPtr) GetObjectPtr (Main_WaitmessageField);
          StrCopy (WaitmessageString, "Please be patient. Tests will take approx. 20 seconds to perform.");
          FldInsert (ActiveField, WaitmessageString, StrLen(WaitmessageString));
          IntMarks   = IntMark();
          MemMarks   = MemMark();
          FloatMarks = FloatMark();
          DBMarks    = DBMark();
          FrmGotoForm (GraphMarkForm);
         
        break;
      }
      handled = true;
    break;

    case menuEvent:	      // a menu was called
      switch (event->data.menu.itemID)
      {
        case Main_HelpMenuAbout:   
          FrmHelp(AboutString);
        break;
        case Main_HelpMenuHelp:
          FrmHelp(GeneralHelpString);
        break;
        case Main_HelpMenuResults:
          FrmHelp(ResultsString);
        break;
      }
      handled = true;
    break;

  }
  CALLBACK_EPILOGUE
  return (handled);
}

/***************************
* GraphMarkFormHandleEvent *
***************************/

static Boolean GraphMarkFormHandleEvent (EventPtr event)
{
  Boolean   handled=false;
  FormPtr	frm;
    
  CALLBACK_PROLOGUE
  switch (event->eType)
  {
    case frmOpenEvent:	   // the form was just opened: draw it
      frm = FrmGetActiveForm();
      FrmDrawForm (frm);
      GraphMarks = GraphMark();
      FrmGotoForm (ResultsForm);
      handled = true;
    break;
  }
  CALLBACK_EPILOGUE
  return (handled);
}

/*************************
* ResultsFormHandleEvent *
*************************/

static Boolean ResultsFormHandleEvent (EventPtr event)
{
  DmOpenRef memoDB;
  Boolean   handled=false;
  FormPtr	frm;
  FieldPtr	ActiveField;
  VoidHand  MemoRecordHandle;
  VoidPtr   MemoRecordPtr;
  UInt      MemoIndex=0;
  DateTimeType DateTime;
  UInt      PalmMarks;
  double    DblPalmMarks;
  char      MemoString[200], DateString[50], TimeString[50];
    
  CALLBACK_PROLOGUE
  switch (event->eType)
  {
    case frmOpenEvent:	   // the form was just opened: draw it
      frm = FrmGetActiveForm();
      FrmDrawForm (frm);

/* Write benchmark results into fields */
      EraseField (Results_IntField);
      EraseField (Results_MemField);
      EraseField (Results_FloatField);
      EraseField (Results_PalmField);
      EraseField (Results_DBField);
      EraseField (Results_GraphField);

      StrIToA ((CharPtr) &IntMarksString, IntMarks);
      ActiveField = (FieldPtr) GetObjectPtr (Results_IntField);
      FldInsert (ActiveField, IntMarksString, StrLen(IntMarksString));
      StrIToA ((CharPtr) &MemMarksString, MemMarks);
      ActiveField = (FieldPtr) GetObjectPtr (Results_MemField);
      FldInsert (ActiveField, MemMarksString, StrLen(MemMarksString));
      StrIToA ((CharPtr) &FloatMarksString, FloatMarks);
      ActiveField = (FieldPtr) GetObjectPtr (Results_FloatField);
      FldInsert (ActiveField, FloatMarksString, StrLen(FloatMarksString));
      StrIToA ((CharPtr) &DBMarksString, DBMarks);
      ActiveField = (FieldPtr) GetObjectPtr (Results_DBField);
      FldInsert (ActiveField, DBMarksString, StrLen(DBMarksString));
      StrIToA ((CharPtr) &GraphMarksString, GraphMarks);
      ActiveField = (FieldPtr) GetObjectPtr (Results_GraphField);
      FldInsert (ActiveField, GraphMarksString, StrLen(GraphMarksString));
      MathLibRInt (MathLibRef, ((IntMarks*70.0 + MemMarks*15.0 + FloatMarks*15.0)/100.0), &DblPalmMarks);
      PalmMarks  = (UInt) DblPalmMarks;
      StrIToA ((CharPtr) &PalmMarksString, PalmMarks);
      ActiveField = (FieldPtr) GetObjectPtr (Results_PalmField);
      FldInsert (ActiveField, PalmMarksString, StrLen(PalmMarksString));
      FldReleaseFocus (ActiveField);

    break;
    
    case ctlSelectEvent: // Button pressed
      switch (event->data.ctlEnter.controlID)
      {
        case Results_SaveButton:
            
/* Prepare results memo string for including in MemoPad */
          TimSecondsToDateTime(TimGetSeconds(), &DateTime);
          DateToAscii(DateTime.month, DateTime.day, DateTime.year, dfDMYWithDots, DateString);
          TimeToAscii (DateTime.hour, DateTime.minute, false, TimeString);      
          StrCopy (MemoString, "PocketMark v0.2.1 - ");
          StrCat (MemoString, DateString);      
          StrCat (MemoString, " ");
          StrCat (MemoString, TimeString);
          StrCat (MemoString, "\n\n");
          StrCat (MemoString, "IntMarks: ");
          StrCat (MemoString, IntMarksString);
          StrCat (MemoString, "\nMemMarks: ");
          StrCat (MemoString, MemMarksString);
          StrCat (MemoString, "\nFloatMarks: ");
          StrCat (MemoString, FloatMarksString);
          StrCat (MemoString, "\n\nPALMMARKS: ");
          StrCat (MemoString, PalmMarksString);
          StrCat (MemoString, "\n\nDBMarks: ");
          StrCat (MemoString, DBMarksString);
          StrCat (MemoString, "\nGraphMarks: ");
          StrCat (MemoString, GraphMarksString);
          StrCat (MemoString, "\n");

/* Writing PocketMark results in a MemoMad entry */      
          memoDB = DmOpenDatabaseByTypeCreator('DATA','memo',dmModeReadWrite);
          MemoRecordHandle = DmNewRecord(memoDB,&MemoIndex,StrLen(MemoString)+1);
          MemoRecordPtr = MemHandleLock(MemoRecordHandle);    
          DmSet(MemoRecordPtr,0,StrLen(MemoString)+1,0);
          DmWrite(MemoRecordPtr,0,MemoString,StrLen(MemoString));
          MemHandleUnlock(MemoRecordHandle);
          DmReleaseRecord(memoDB,MemoIndex,true);
          DmCloseDatabase(memoDB);
          
        break;
        }
        
    case menuEvent:	      // a menu was called
      switch (event->data.menu.itemID)
      {
        case Main_HelpMenuAbout:   
          FrmHelp(AboutString);
        break;
        case Main_HelpMenuHelp:
          FrmHelp(GeneralHelpString);
        break;
        case Main_HelpMenuResults:
          FrmHelp(ResultsString);
        break;
      }
    break;
    
  handled = true;
  }
  CALLBACK_EPILOGUE
  return (handled);
}

/*******************************
** ApplicationHandleEvent **
*******************************/

static Boolean ApplicationHandleEvent (EventPtr event)
{                       // Note: this handler is mandatory !
  FormPtr	frm;
  Boolean	handled=false;
  Int		   formID;

  CALLBACK_PROLOGUE  
  if (event->eType == frmLoadEvent)
  {
    formID  = event->data.frmLoad.formID;
    frm     = FrmInitForm (formID);
    FrmSetActiveForm (frm);
    switch (formID)	   // setting the right handler for requested form
    {
      case MainForm:	
        FrmSetEventHandler (frm, MainFormHandleEvent);
      break;
      case GraphMarkForm:	
        FrmSetEventHandler (frm, GraphMarkFormHandleEvent);
      break;
      case ResultsForm:	
        FrmSetEventHandler (frm, ResultsFormHandleEvent);
      break;
    }
    handled = true;
  }
  CALLBACK_EPILOGUE
  return (handled);
}

/***********************
** EventLoop          **
***********************/

static void EventLoop (void)
{
  EventType	eventtype;
  Word 		error;

  do
  {
    EvtGetEvent (&eventtype, evtWaitForever);		   // wait for next event, no timeout
    if (!SysHandleEvent (&eventtype))			         // 1^ try: system event ?
      if (!MenuHandleEvent (NULL,&eventtype,&error))	// 2^ try: menu event ?
        if (!ApplicationHandleEvent (&eventtype))  	// 3^ try: application change ?
          FrmDispatchEvent (&eventtype);		         // OK, so the event was for a form in this application
  }
  while (eventtype.eType != appStopEvent);
}

/***********************
** PilotMain          **
***********************/

DWord PilotMain (Word launchCode, Ptr paramBlock, Word launchFlags)
{
  if (launchCode == sysAppLaunchCmdNormalLaunch)	   // we only consider a "normal launch" case
    if (!StartApplication())			
    {
      EventLoop();
      StopApplication();
    }
  return (0);
}
